@@ -15,17 +15,26 @@ import queue
 import time
 import kthread
 
+# Those three modules are required to retrieve cert from the router and extract cn name
+import ssl
+from cryptography import x509
+from cryptography.hazmat.backends import default_backend
+
 # Own modules
 from pygnmi.path_generator import gnmi_path_generator
 
 
+# Logger
+logger = logging.getLogger(__name__)
+logger.addHandler(logging.StreamHandler(sys.stdout))
+
 # Classes
 class gNMIclient(object):
     """
     This class instantiates the object, which interacts with the network elements over gNMI.
     """
     def __init__(self, target: tuple, username: str = None, password: str = None, 
-                 debug: bool = False, insecure: bool = False, path_cert: str = None, override: str = None,
+                 debug: bool = False, insecure: bool = False, path_cert: str = None, path_key: str = None, path_root: str = None, override: str = None,
                  gnmi_timeout: int = 5):
 
         """
@@ -36,36 +45,117 @@ class gNMIclient(object):
         self.__debug = debug
         self.__insecure = insecure
         self.__path_cert = path_cert
+        self.__path_key = path_key
+        self.__path_root = path_root
         self.__options=[('grpc.ssl_target_name_override', override)] if override else None
         self.__gnmi_timeout = gnmi_timeout
 
-        if re.match('.*:.*', target[0]):
+        self.__target_path = f'{target[0]}:{target[1]}'
+        if re.match('unix:.*', target[0]):
+            self.__target = target
+            self.__target_path = target[0]
+        elif re.match('.*:.*', target[0]):
             self.__target = (f'[{target[0]}]', target[1])
         else:
             self.__target = target
 
-
     def __enter__(self):
         """
         Building the connectivity towards network element over gNMI
         """
 
         if self.__insecure:
-            self.__channel = grpc.insecure_channel(f'{self.__target[0]}:{self.__target[1]}', self.__metadata)
+            self.__channel = grpc.insecure_channel(self.__target_path, self.__metadata)
             grpc.channel_ready_future(self.__channel).result(timeout=self.__gnmi_timeout)
             self.__stub = gNMIStub(self.__channel)
 
         else:
-            if self.__path_cert:
+            if self.__path_cert and self.__path_key and self.__path_root:
+                try:
+                    cert = open(self.__path_cert, 'rb').read()
+                    key = open(self.__path_key, 'rb').read()
+                    root_cert = open(self.__path_root, 'rb').read()
+                    cert = grpc.ssl_channel_credentials(root_certificates=root_cert, private_key=key, certificate_chain=cert)
+                except:
+                    logging.error('The SSL certificate cannot be opened.')
+                    raise Exception('The SSL certificate cannot be opened.')
+
+            elif self.__path_cert:
                 try:
                     with open(self.__path_cert, 'rb') as f:
                         cert = grpc.ssl_channel_credentials(f.read())
 
+                except:
+                    logger.error('The SSL certificate cannot be opened.')
+                    raise Exception('The SSL certificate cannot be opened.')
+                    
+            else:
+                try:
+                    ssl_cert = ssl.get_server_certificate((self.__target[0], self.__target[1])).encode("utf-8")
+                    ssl_cert_deserialized = x509.load_pem_x509_certificate(ssl_cert, default_backend())
+                    ssl_cert_common_names = ssl_cert_deserialized.subject.get_attributes_for_oid(x509.oid.NameOID.COMMON_NAME)
+                    ssl_target_name_override = ssl_cert_common_names[0].value
+                    self.__options = [("grpc.ssl_target_name_override", ssl_target_name_override)]
+                    logger.warning('ssl_target_name_override is applied, should be used for testing only!')
+                    cert = grpc.ssl_channel_credentials(ssl_cert)
+
+                except:
+                    logger.error(f'The SSL certificate cannot be retrieved from {self.__target}')
+                    raise Exception(f'The SSL certificate cannot be retrieved from {self.__target}') 
+
+            self.__channel = grpc.secure_channel(self.__target_path, 
+                                                 credentials=cert, options=self.__options)
+            grpc.channel_ready_future(self.__channel).result(timeout=self.__gnmi_timeout)
+            self.__stub = gNMIStub(self.__channel)
+
+        return self
+
+
+    def connect(self):
+        """
+        Building the connectivity towards network element over gNMI without conext manage with ... as ...
+        """
+
+        if self.__insecure:
+            self.__channel = grpc.insecure_channel(self.__target_path, self.__metadata)
+            grpc.channel_ready_future(self.__channel).result(timeout=self.__gnmi_timeout)
+            self.__stub = gNMIStub(self.__channel)
+
+        else:
+            if self.__path_cert and self.__path_key and self.__path_root:
+                try:
+                    cert = open(self.__path_cert, 'rb').read()
+                    key = open(self.__path_key, 'rb').read()
+                    root_cert = open(self.__path_root, 'rb').read()
+                    cert = grpc.ssl_channel_credentials(root_certificates=root_cert, private_key=key, certificate_chain=cert)
                 except:
                     logging.error('The SSL certificate cannot be opened.')
                     raise Exception('The SSL certificate cannot be opened.')
 
-            self.__channel = grpc.secure_channel(f'{self.__target[0]}:{self.__target[1]}', 
+            elif self.__path_cert:
+                try:
+                    with open(self.__path_cert, 'rb') as f:
+                        cert = grpc.ssl_channel_credentials(f.read())
+
+                except:
+                    logger.error('The SSL certificate cannot be opened.')
+                    raise Exception('The SSL certificate cannot be opened.')
+                    
+            else:
+                try:
+                    ssl_cert = ssl.get_server_certificate((self.__target[0], self.__target[1])).encode("utf-8")
+                    ssl_cert_deserialized = x509.load_pem_x509_certificate(ssl_cert, default_backend())
+                    ssl_cert_common_names = ssl_cert_deserialized.subject.get_attributes_for_oid(x509.oid.NameOID.COMMON_NAME)
+                    ssl_target_name_override = ssl_cert_common_names[0].value
+                    self.__options = [("grpc.ssl_target_name_override", ssl_target_name_override)]
+                    logger.warning('ssl_target_name_override is applied, should be used for testing only!')
+                    cert = grpc.ssl_channel_credentials(ssl_cert)
+
+                except:
+                    logger.error(f'The SSL certificate cannot be retrieved from {self.__target}')
+                    raise Exception(f'The SSL certificate cannot be retrieved from {self.__target}') 
+
+            self.__channel = grpc.secure_channel(self.__target_path, 
                                                  credentials=cert, options=self.__options)
             grpc.channel_ready_future(self.__channel).result(timeout=self.__gnmi_timeout)
             self.__stub = gNMIStub(self.__channel)
@@ -78,7 +168,7 @@ class gNMIclient(object):
         Collecting the gNMI capabilities of the network device.
         There are no arguments needed for this call
         """
-        logging.info(f'Collecting Capabilities...')
+        logger.info(f'Collecting Capabilities...')
 
         try:
             gnmi_message_request = CapabilityRequest()
@@ -124,18 +214,19 @@ class gNMIclient(object):
                 if gnmi_message_response.gNMI_version:
                     response.update({'gnmi_version': gnmi_message_response.gNMI_version})
 
-            logging.info(f'Collection of Capabilities is successfull')
+            logger.info(f'Collection of Capabilities is successfull')
 
             self.__capabilities = response
             return response
 
         except grpc._channel._InactiveRpcError as err:
-            print(f"Host: {self.__target[0]}:{self.__target[1]}\nError: {err.details()}")
-            logging.critical(f"GRPC ERROR Host: {self.__target[0]}:{self.__target[1]}, Error: {err.details()}")
+            print(f"Host: {self.__target_path}\nError: {err.details()}")
+            logger.critical(f"GRPC ERROR Host: {self.__target_path}, Error: {err.details()}")
+
             raise Exception (err)
 
         except:
-            logging.error(f'Collection of Capabilities is failed.')
+            logger.error(f'Collection of Capabilities is failed.')
 
             return None
 
@@ -167,7 +258,7 @@ class gNMIclient(object):
           - ascii
           - json_ietf
         """
-        logging.info(f'Collecting info from requested paths (Get operation)...')
+        logger.info(f'Collecting info from requested paths (Get operation)...')
 
         datatype = datatype.lower()
         type_dict = {'all', 'config', 'state', 'operational'}
@@ -183,7 +274,7 @@ class gNMIclient(object):
             elif datatype == 'operational':
                 pb_datatype = 3
             else:
-                logging.error('The GetRequst data type is not within the dfined range')
+                logger.error('The GetRequst data type is not within the dfined range')
 
         if encoding in encoding_set:
             if encoding.lower() == 'json':
@@ -205,7 +296,7 @@ class gNMIclient(object):
                 protobuf_paths = [gnmi_path_generator(pe) for pe in path]
 
         except:
-            logging.error(f'Conversion of gNMI paths to the Protobuf format failed')
+            logger.error(f'Conversion of gNMI paths to the Protobuf format failed')
             raise Exception ('Conversion of gNMI paths to the Protobuf format failed')
 
         if self.__capabilities and 'supported_encodings' in self.__capabilities:
@@ -287,12 +378,13 @@ class gNMIclient(object):
             return response
 
         except grpc._channel._InactiveRpcError as err:
-            print(f"Host: {self.__target[0]}:{self.__target[1]}\nError: {err.details()}")
-            logging.critical(f"GRPC ERROR Host: {self.__target[0]}:{self.__target[1]}, Error: {err.details()}")
+            print(f"Host: {self.__target_path}\nError: {err.details()}")
+            logger.critical(f"GRPC ERROR Host: {self.__target_path}, Error: {err.details()}")
+
             raise Exception (err)
 
         except:
-            logging.error(f'Collection of Get information failed is failed.')
+            logger.error(f'Collection of Get information failed is failed.')
 
             return None
 
@@ -326,7 +418,7 @@ class gNMIclient(object):
         encoding_set = {'json', 'bytes', 'proto', 'ascii', 'json_ietf'}
 
         if encoding not in encoding_set:
-            logging.error(f'The encoding {encoding} is not supported. The allowed are: {", ".join(encoding_set)}.')
+            logger.error(f'The encoding {encoding} is not supported. The allowed are: {", ".join(encoding_set)}.')
             raise Exception (f'The encoding {encoding} is not supported. The allowed are: {", ".join(encoding_set)}.')
 
         if delete:
@@ -335,11 +427,11 @@ class gNMIclient(object):
                     del_protobuf_paths = [gnmi_path_generator(pe) for pe in delete]
 
                 except:
-                    logging.error(f'Conversion of gNMI paths to the Protobuf format failed')
+                    logger.error(f'Conversion of gNMI paths to the Protobuf format failed')
                     raise Exception (f'Conversion of gNMI paths to the Protobuf format failed')
 
             else:
-                logging.error(f'The provided input for Set message (delete operation) is not list.')
+                logger.error(f'The provided input for Set message (delete operation) is not list.')
                 raise Exception (f'The provided input for Set message (delete operation) is not list.')
 
         if replace:
@@ -361,11 +453,11 @@ class gNMIclient(object):
                             replace_msg.append(Update(path=u_path, val=TypedValue(json_ietf_val=u_val)))
 
                     else:
-                        logging.error(f'The input element for Update message must be tuple, got {ue}.')
+                        logger.error(f'The input element for Update message must be tuple, got {ue}.')
                         raise Exception (f'The input element for Update message must be tuple, got {ue}.')
 
             else:
-                logging.error(f'The provided input for Set message (replace operation) is not list.')
+                logger.error(f'The provided input for Set message (replace operation) is not list.')
                 raise Exception ('The provided input for Set message (replace operation) is not list.')
 
         if update:
@@ -387,11 +479,11 @@ class gNMIclient(object):
                             update_msg.append(Update(path=u_path, val=TypedValue(json_ietf_val=u_val)))
 
                     else:
-                        logging.error(f'The input element for Update message must be tuple, got {ue}.')
+                        logger.error(f'The input element for Update message must be tuple, got {ue}.')
                         raise Exception (f'The input element for Update message must be tuple, got {ue}.')
 
             else:
-                logging.error(f'The provided input for Set message (update operation) is not list.')
+                logger.error(f'The provided input for Set message (update operation) is not list.')
                 raise Exception ('The provided input for Set message (replace operation) is not list.')
 
         try:
@@ -453,16 +545,17 @@ class gNMIclient(object):
                 return response
 
             else:
-                logging.error('Failed parsing the SetResponse.')
+                logger.error('Failed parsing the SetResponse.')
                 return None
 
         except grpc._channel._InactiveRpcError as err:
-            print(f"Host: {self.__target[0]}:{self.__target[1]}\nError: {err.details()}")
-            logging.critical(f"GRPC ERROR Host: {self.__target[0]}:{self.__target[1]}, Error: {err.details()}")
+            print(f"Host: {self.__target_path}\nError: {err.details()}")
+            logger.critical(f"GRPC ERROR Host: {self.__target_path}, Error: {err.details()}")
+
             raise Exception (err)
 
         except:
-            logging.error(f'Collection of Set information failed is failed.')
+            logger.error(f'Collection of Set information failed is failed.')
             return None
 
 
@@ -470,7 +563,7 @@ class gNMIclient(object):
         """
         Implentation of the subsrcibe gNMI RPC to pool
         """
-        logging.info(f'Collecting Telemetry...')
+        logger.info(f'Collecting Telemetry...')
 
         if (subscribe and poll) or (subscribe and aliases) or (poll and aliases):
             raise Exception('Subscribe request supports only one request at a time.')
@@ -483,7 +576,7 @@ class gNMIclient(object):
                     gnmi_message_request = SubscribeRequest(poll=request)
 
             else:
-                logging.error('Subscribe pool request is specificed, but the value is not boolean.')
+                logger.error('Subscribe pool request is specificed, but the value is not boolean.')
 
         if aliases:
             if isinstance(aliases, list):
@@ -499,7 +592,7 @@ class gNMIclient(object):
                 gnmi_message_request = SubscribeRequest(aliases=request)
 
             else:
-                logging.error('Subscribe aliases request is specified, but the value is not list.')
+                logger.error('Subscribe aliases request is specified, but the value is not list.')
 
         if subscribe:
             if isinstance(subscribe, dict):
@@ -633,7 +726,7 @@ class gNMIclient(object):
                 gnmi_message_request = SubscribeRequest(subscribe=request)
 
             else:
-                logging.error('Subscribe subscribe requst is specified, but the value is not list.')
+                logger.error('Subscribe subscribe requst is specified, but the value is not list.')
 
             if self.__debug:
                 print("gNMI request:\n------------------------------------------------")
@@ -757,50 +850,7 @@ def telemetryParser(in_message=None, debug: bool = False):
             return response
 
     except:
-        logging.error(f'Parsing of telemetry information is failed.')
+        logger.error(f'Parsing of telemetry information is failed.')
 
         return None
 
-
-class Nonblocking_iterator:
-    def __init__(self, blocking_generator, timeout, stop_if_timeout = False):
-        self.timeout = timeout
-        self.stop_if_timeout = stop_if_timeout
-        self.msg_queue = queue.Queue()
-        self.blocking_generator = blocking_generator
-        self._thread = kthread.KThread(
-            target=self._put_messages_from_blocking_generator_to_queue)
-        self._thread.start()
-
-    def _put_messages_from_blocking_generator_to_queue(self):
-        while True:
-            self.msg_queue.put(next(self.blocking_generator))
-
-    def terminate(self):
-        self._thread.terminate()
-        while self._thread.is_alive():
-            self._thread.join()
-        time.sleep(1)
-
-    def is_alive(self):
-        return self._thread.is_alive()
-
-    def join(self):
-        self._thread.join()
-
-    def __iter__(self):
-        return self
-
-    def __next__(self):
-        try:
-            return self.msg_queue.get(timeout=self.timeout)
-        except queue.Empty:
-            self._thread.terminate()
-            while self._thread.is_alive():
-                self._thread.join()
-            time.sleep(1)
-            raise Exception("timeout error")
-
-    def next(self):
-        return self.__next__()
-
